@@ -72,6 +72,11 @@ struct msg_buff {
 
 #define DEFAULT_MSG_SIZE	8192
 
+static inline unsigned char *msg_tail_pointer(struct msg_buff *msg)
+{
+	return msg->tail;
+}
+
 static inline int msg_tailroom(struct msg_buff *msg)
 {
 	return msg->end - msg->tail;
@@ -333,6 +338,9 @@ extern int		nla_put_nohdr(struct msg_buff *msg, int attrlen,
 				      const void *data);
 extern int		nla_append(struct msg_buff *msg, int attrlen,
 				   const void *data);
+extern int		nla_put_64bit(struct msg_buff *msg, int attrtype,
+				int attrlen, const void *data, int padattr);
+#define COMPAT_HAVE_NLA_PUT_64BIT 1
 
 /**************************************************************************
  * Netlink Messages
