@@ -24,6 +24,12 @@
 	.nla_policy = p ## _nl_policy,							\
 	.nla_policy_size = ARRAY_SIZE(p ## _nl_policy)
 
+
+struct en_map {
+	const char *name;
+	int value;
+};
+
 /* ============================================================================================== */
 
 static int enum_string_to_int(const char **map, int size, const char *value,
@@ -341,6 +347,158 @@ struct field_class fc_numeric = {
 
 /* ---------------------------------------------------------------------------------------------- */
 
+static int enum_num_to_int(const struct en_map *map, int map_size, const char *value,
+	enum new_strtoll_errs *err_p)
+{
+	enum new_strtoll_errs e;
+	unsigned long long l;
+	int i;
+
+	if (!value) {
+		if (err_p)
+			*err_p = MSE_MISSING_NUMBER;
+		return -1;
+	}
+
+	for (i = 0; i < map_size; i++) {
+		if (!strcmp(value, map[i].name))
+			return map[i].value;
+	}
+
+	e = new_strtoll(value, 1, &l);
+	if (err_p)
+		*err_p = e;
+	return e == MSE_OK ? l : -1;
+}
+
+static bool enum_num_is_default(struct field_def *field, const char *value)
+{
+	int n;
+
+	n = enum_num_to_int(field->u.en.map, field->u.en.map_size, value, NULL);
+	return n == field->u.en.def;
+}
+
+static bool enum_num_is_equal(struct field_def *field, const char *a, const char *b)
+{
+	return !strcmp(a, b);
+}
+
+static const char *enum_num_to_string(struct field_def *field, int value)
+{
+	static char buffer[1 + 10 + 2];
+	int i;
+
+	for (i = 0; i < field->u.en.map_size; i++) {
+		if (value == field->u.en.map[i].value)
+			return field->u.en.map[i].name;
+	}
+
+	i = snprintf(buffer, sizeof(buffer), "%d", value);
+	assert(i < sizeof(buffer));
+
+	return buffer;
+}
+
+static const char *get_enum_num(struct context_def *ctx, struct field_def *field, struct nlattr *nla)
+{
+	int n;
+
+	assert(type_of_field(ctx, field) == NLA_U32);
+	n = nla_get_u32(nla);
+
+	return enum_num_to_string(field, n);
+}
+
+static bool put_enum_num(struct context_def *ctx, struct field_def *field,
+			struct msg_buff *msg, const char *value)
+{
+	int n;
+
+	n = enum_num_to_int(field->u.en.map, field->u.en.map_size, value, NULL);
+	if (n == -1)
+		return false;
+	assert(type_of_field(ctx, field) == NLA_U32);
+	nla_put_u32(msg, field->nla_type, n);
+	return true;
+}
+
+static int enum_num_usage(struct field_def *field, char *str, int size)
+{
+	const struct en_map *map = field->u.en.map;
+	char sep = '{';
+	int i, len = 0, l;
+
+	l = snprintf(str, size, "[--%s=", field->name);
+	len += l; size -= l;
+	for (i = 0; i < field->u.en.map_size; i++) {
+		l = snprintf(str + len, size, "%c%s", sep, map[i].name);
+		len += l; size -= l;
+		sep = '|';
+	}
+	assert (sep != '{');
+	l = snprintf(str + len, size, "}|(%d ... %d)]",
+		     field->u.en.min, field->u.en.max);
+	len += l;
+	/* size -= l; */
+	return len;
+}
+
+static void enum_num_describe_xml(struct field_def *field)
+{
+	const struct en_map *map = field->u.en.map;
+	int i;
+
+	printf("\t<option name=\"%s\" type=\"numeric-or-symbol\">\n"
+	       "\t\t<min>%d</min>\n"
+	       "\t\t<max>%d</max>\n"
+	       "\t\t<default>%s</default>\n",
+	       field->name,
+	       field->u.en.min,
+	       field->u.en.max,
+	       enum_num_to_string(field, field->u.en.def));
+
+	for (i = 0; i < field->u.en.map_size; i++)
+		printf("\t\t<symbol>%s</symbol>\n", map[i].name);
+
+	printf("\t</option>\n");
+}
+
+static enum check_codes enum_num_check(struct field_def *field, const char *value)
+{
+	enum new_strtoll_errs e = 777;
+	int n;
+
+	n = enum_num_to_int(field->u.en.map, field->u.en.map_size, value, &e);
+
+	if (n == -1 && e != MSE_OK)
+		return CC_NOT_AN_ENUM_NUM;
+
+	/* n positive, but e not touched -> it was a symbolic value, no range check */
+	if (e == 777)
+		return CC_OK;
+
+	if (n < field->u.en.min)
+		return CC_TOO_SMALL;
+
+	if (n > field->u.en.max)
+		return CC_TOO_BIG;
+
+	return CC_OK;
+}
+
+struct field_class fc_enum_num = {
+	.is_default = enum_num_is_default,
+	.is_equal = enum_num_is_equal,
+	.get = get_enum_num,
+	.put = put_enum_num,
+	.usage = enum_num_usage,
+	.describe_xml = enum_num_describe_xml,
+	.check = enum_num_check,
+};
+
+/* ---------------------------------------------------------------------------------------------- */
+
 static int boolean_string_to_int(const char *value)
 {
 	if (!value || !strcmp(value, "yes"))
@@ -583,6 +741,16 @@ struct field_class fc_string = {
 	STRING(f),									\
 	.u = { .s = { .max_len = l } } }
 
+#define ENUM_NUM(f, d, num_min, num_max)			\
+	.nla_type = T_ ## f,					\
+	.ops = &fc_enum_num,					\
+	.u = { .en = {						\
+		.map = f ## _map,				\
+		.map_size = ARRAY_SIZE(f ## _map),		\
+		.min = num_min,					\
+		.max = num_max,					\
+		.def = DRBD_ ## d ## _DEF, } }			\
+
 /* ============================================================================================== */
 
 const char *wire_protocol_map[] = {
@@ -634,10 +802,13 @@ const char *rr_conflict_map[] = {
 };
 
 const char *on_no_data_map[] = {
-	[OND_IO_ERROR]		= "io-error",
-	[OND_SUSPEND_IO]	= "suspend-io",
+	[OND_IO_ERROR] = "io-error",
+	[OND_SUSPEND_IO] = "suspend-io",
 };
 
+#define on_no_quorum_map on_no_data_map
+/* ONQ_XX == OND_XX */
+
 const char *on_congestion_map[] = {
 	[OC_BLOCK] = "block",
 	[OC_PULL_AHEAD] = "pull-ahead",
@@ -658,6 +829,14 @@ const char *read_balancing_map[] = {
 	[RB_1M_STRIPING] = "1M-striping"
 };
 
+const struct en_map quorum_map[] = {
+	{ "off", QOU_OFF },
+	{ "majority", QOU_MAJORITY },
+	{ "all", QOU_ALL },
+};
+
+#define quorum_min_redundancy_map quorum_map
+
 #define CHANGEABLE_DISK_OPTIONS								\
 	{ "on-io-error", ENUM(on_io_error, ON_IO_ERROR) },				\
 	/*{ "fencing", ENUM(fencing_policy, FENCING) },*/				\
@@ -670,6 +849,8 @@ const char *read_balancing_map[] = {
 	{ "al-updates", BOOLEAN(al_updates, AL_UPDATES) },				\
 	{ "discard-zeroes-if-aligned",							\
 		BOOLEAN(discard_zeroes_if_aligned, DISCARD_ZEROES_IF_ALIGNED) },	\
+	{ "disable-write-same",								\
+		BOOLEAN(disable_write_same, DISABLE_WRITE_SAME) },			\
 	{ "disk-timeout", NUMERIC(disk_timeout,	DISK_TIMEOUT),				\
 	  .unit = "1/10 seconds" },							\
 	{ "read-balancing", ENUM(read_balancing, READ_BALANCING) },			\
@@ -759,6 +940,7 @@ struct context_def detach_cmd_ctx = {
 	.nla_type = DRBD_NLA_DETACH_PARMS,
 	.fields = {
 		{ "force", FLAG(force_detach) },
+		{ "diskless", FLAG(intentional_diskless_detach) },
 		{ }
 	},
 };
@@ -836,6 +1018,9 @@ struct context_def resource_options_ctx = {
 		{ "auto-promote-timeout", NUMERIC(auto_promote_timeout, AUTO_PROMOTE_TIMEOUT),
 		  .unit = "1/10 seconds"},
 		{ "max-io-depth", NUMERIC(nr_requests, NR_REQUESTS) },
+		{ "quorum", ENUM_NUM(quorum, QUORUM, 1, DRBD_PEERS_MAX) },
+		{ "on-no-quorum", ENUM(on_no_quorum, ON_NO_QUORUM) },
+		{ "quorum-minimum-redundancy", ENUM_NUM(quorum_min_redundancy, QUORUM, 1, DRBD_PEERS_MAX) },
 		{ } },
 };
 
@@ -863,6 +1048,7 @@ struct context_def device_options_ctx = {
 	.nla_type = DRBD_NLA_DEVICE_CONF,
 	.fields = {
 		{ "max-bio-size", NUMERIC(max_bio_size, MAX_BIO_SIZE) },
+		{ "diskless", FLAG(intentional_diskless) },
 		{ } },
 };
 
@@ -899,6 +1085,12 @@ struct context_def create_md_ctx = {
 		{ } },
 };
 
+struct context_def dump_md_ctx = {
+       .fields = {
+		{ .name = "force", .argument_is_optional = true },
+		{ } },
+};
+
 struct context_def adjust_ctx = {
 	.fields = {
 		{ "skip-disk", .argument_is_optional = true },
@@ -923,6 +1115,7 @@ struct context_def handlers_ctx = {
 		{ "after-resync-target", .ops = &fc_string, .needs_double_quoting = true},
 		{ "before-resync-source", .ops = &fc_string, .needs_double_quoting = true},
 		{ "out-of-sync", .ops = &fc_string, .needs_double_quoting = true},
+		{ "quorum-lost", .ops = &fc_string, .needs_double_quoting = true},
 		{ } },
 };
 
