@@ -276,12 +276,9 @@ void check_upr_init(void)
 	created = 1;
 }
 
-/* FIXME
- * strictly speaking we don't need to check for uniqueness of disk and device names,
- * but for uniqueness of their major:minor numbers ;-)
- */
-
-int vcheck_uniq(void **bt, const char *what, const char *fmt, va_list ap)
+int vcheck_uniq_file_line(
+	const char *file, const int line,
+	void **bt, const char *what, const char *fmt, va_list ap)
 {
 	int rv;
 	ENTRY *e, *f;
@@ -308,7 +305,7 @@ int vcheck_uniq(void **bt, const char *what, const char *fmt, va_list ap)
 		err("Oops, unset argument in %s:%d.\n", __FILE__, __LINE__);
 		exit(E_THINKO);
 	}
-	m_asprintf((char **)&e->data, "%s:%u", config_file, fline);
+	m_asprintf((char **)&e->data, "%s:%u", file, line);
 
 	f = tfind(e, bt, btree_key_cmp);
 	if (f) {
@@ -334,6 +331,11 @@ int vcheck_uniq(void **bt, const char *what, const char *fmt, va_list ap)
 	return !f;
 }
 
+int vcheck_uniq(void **bt, const char *what, const char *fmt, va_list ap)
+{
+	return vcheck_uniq_file_line(config_file, fline, bt, what, fmt, ap);
+}
+
 static void pe_expected(const char *exp)
 {
 	const char *s = yytext;
@@ -399,6 +401,9 @@ static void parse_global(void)
 		int token = yylex();
 		fline = line;
 		switch (token) {
+		case TK_UDEV_ALWAYS_USE_VNR:
+			global_options.udev_always_symlink_vnr = 1;
+			break;
 		case TK_DISABLE_IP_VERIFICATION:
 			global_options.disable_ip_verification = 1;
 			break;
@@ -503,6 +508,7 @@ static void pe_field(struct field_def *field, enum check_codes e, char *value)
 		[CC_TOO_SMALL] = "too small",
 		[CC_TOO_BIG] = "too big",
 		[CC_STR_TOO_LONG] = "too long",
+		[CC_NOT_AN_ENUM_NUM] = "not valid",
 	};
 	err("%s:%u: Parse error: while parsing value ('%.20s%s')\nfor %s. Value is %s.\n",
 	    config_file, line,
@@ -511,8 +517,10 @@ static void pe_field(struct field_def *field, enum check_codes e, char *value)
 
 	if (e == CC_NOT_AN_ENUM)
 		pe_valid_enums(field->u.e.map, field->u.e.size);
-	if (e == CC_STR_TOO_LONG)
+	else if (e == CC_STR_TOO_LONG)
 		err("max len: %u\n", field->u.s.max_len - 1);
+	/* else if (e == CC_NOT_AN_ENUM_NUM)
+	   pe_valid_enum_num((field->u.en.map, field->u.en.map_size); */
 
 	if (config_valid <= 1)
 		config_valid = 0;
@@ -907,9 +915,11 @@ out:
 		return;
 
 	STAILQ_FOREACH(h, on_hosts, link) {
-		check_uniq("device-minor", "device-minor:%s:%u", h->name, vol->device_minor);
+		check_uniq_file_line(vol->v_config_file, vol->v_device_line,
+			"device-minor", "device-minor:%s:%u", h->name, vol->device_minor);
 		if (vol->device)
-			check_uniq("device", "device:%s:%s", h->name, vol->device);
+			check_uniq_file_line(vol->v_config_file, vol->v_device_line,
+				"device", "device:%s:%s", h->name, vol->device);
 	}
 }
 
@@ -945,7 +955,7 @@ struct d_volume *volume0(struct volumes *volumes)
 			return vol;
 
 		config_valid = 0;
-		err("%s:%d: Explicit and implicit volumes not allowed\n",
+		err("%s:%d: mixing explicit and implicit volumes is not allowed\n",
 		    config_file, line);
 		return vol;
 	}
@@ -953,6 +963,9 @@ struct d_volume *volume0(struct volumes *volumes)
 
 int parse_volume_stmt(struct d_volume *vol, struct names* on_hosts, int token)
 {
+	if (!vol->v_config_file)
+		vol->v_config_file = config_file;
+
 	switch (token) {
 	case TK_DISK:
 		token = yylex();
@@ -969,14 +982,17 @@ int parse_volume_stmt(struct d_volume *vol, struct names* on_hosts, int token)
 			pe_expected_got( "TK_STRING | {", token);
 		}
 		vol->parsed_disk = 1;
+		vol->v_disk_line = fline;
 		break;
 	case TK_DEVICE:
 		parse_device(on_hosts, vol);
 		vol->parsed_device = 1;
+		vol->v_device_line = fline;
 		break;
 	case TK_META_DISK:
 		parse_meta_disk(vol);
 		vol->parsed_meta_disk = 1;
+		vol->v_meta_disk_line = fline;
 		break;
 	case TK_FLEX_META_DISK:
 		EXP(TK_STRING);
@@ -990,6 +1006,7 @@ int parse_volume_stmt(struct d_volume *vol, struct names* on_hosts, int token)
 		}
 		EXP(';');
 		vol->parsed_meta_disk = 1;
+		vol->v_meta_disk_line = fline;
 		break;
 	case TK_SKIP:
 		parse_skip();
@@ -1365,7 +1382,10 @@ int parse_proxy_options_section(struct d_proxy_info **pp)
 	int token;
 	struct d_proxy_info *proxy;
 
-	proxy = *pp ? *pp : calloc(1, sizeof(struct d_proxy_info));
+	if (!*pp)
+		*pp = calloc(1, sizeof(struct d_proxy_info));
+	proxy = *pp;
+
 	token = yylex();
 	if (token != TK_PROXY) {
 		yyrestart(yyin); /* flushes flex's buffers */
@@ -1647,6 +1667,11 @@ static struct connection *parse_connection(enum pr_flags flags)
 			insert_tail(&conn->paths, parse_path());
 			break;
 		case '}':
+			if (STAILQ_EMPTY(&conn->paths) && !(flags & PARSE_FOR_ADJUST)) {
+				err("%s:%d: connection without a single path (maybe empty?) not allowed\n",
+						config_file, fline);
+				config_valid = 0;
+			}
 			return conn;
 		default:
 			pe_expected_got( "host | net | skip | }", token);
@@ -1752,9 +1777,12 @@ struct d_resource* parse_resource(char* res_name, enum pr_flags flags)
 			break;
 		case TK_DISK:
 			switch (token=yylex()) {
-			case TK_STRING:
+			case TK_STRING:{
 				/* open coded parse_volume_stmt() */
-				volume0(&res->volumes)->disk = yylval.txt;
+				struct d_volume *vol = volume0(&res->volumes);
+				vol->disk = yylval.txt;
+				vol->parsed_disk = 1;
+				}
 				EXP(';');
 				break;
 			case '{':
@@ -1764,7 +1792,7 @@ struct d_resource* parse_resource(char* res_name, enum pr_flags flags)
 				break;
 			default:
 				check_string_error(token);
-				pe_expected_got( "TK_STRING | {", token);
+				pe_expected_got("TK_STRING | {", token);
 			}
 			break;
 		case TK_NET:
@@ -1980,7 +2008,7 @@ void include_stmt(char *str)
 				fclose(f);
 			} else {
 				err("%s:%d: Failed to open include file '%s'.\n",
-				    config_save, line, yylval.txt);
+				    config_save, line, glob_buf.gl_pathv[i]);
 				config_valid = 0;
 			}
 		}
@@ -1988,7 +2016,7 @@ void include_stmt(char *str)
 	} else if (r == GLOB_NOMATCH) {
 		if (!strchr(str, '?') && !strchr(str, '*') && !strchr(str, '[')) {
 			err("%s:%d: Failed to open include file '%s'.\n",
-			    config_save, line, yylval.txt);
+			    config_save, line, str);
 			config_valid = 0;
 		}
 	} else {
@@ -2050,6 +2078,18 @@ int check_uniq(const char *what, const char *fmt, ...)
 	return rv;
 }
 
+int check_uniq_file_line(const char *file, const int line, const char *what, const char *fmt, ...)
+{
+	int rv;
+	va_list ap;
+
+	va_start(ap, fmt);
+	rv = vcheck_uniq_file_line(file, line, &global_btree, what, fmt, ap);
+	va_end(ap);
+
+	return rv;
+}
+
 /* unique per resource */
 int check_upr(const char *what, const char *fmt, ...)
 {
