@@ -31,11 +31,15 @@
 #include <assert.h>
 #include "drbdtool_common.h"
 #include "drbdadm.h"
+#include "config_flags.h"
 #include <search.h>
+#include <arpa/inet.h>
+#include <sys/socket.h>
+#include <netinet/in.h>
 
 static void inherit_volumes(struct volumes *from, struct d_host_info *host);
 static void check_volume_sets_equal(struct d_resource *, struct d_host_info *, struct d_host_info *);
-static void expand_opts(struct options *common, struct options *options);
+static void expand_opts(struct d_resource *, struct context_def *, struct options *, struct options *);
 
 static void append_names(struct names *head, struct names *to_copy)
 {
@@ -79,13 +83,14 @@ void set_on_hosts_in_res(struct d_resource *res)
 						struct d_volume *vol;
 
 						for_each_volume(vol, &host->volumes)
-							check_uniq("device-minor", "device-minor:%s:%u", h->name,
+							check_uniq_file_line(vol->v_config_file, vol->v_device_line,
+								"device-minor", "device-minor:%s:%u", h->name,
 									vol->device_minor);
 
 						for_each_volume(vol, &host->volumes)
 							if (vol->device)
-								check_uniq("device", "device:%s:%s", h->name,
-										vol->device);
+								check_uniq_file_line(vol->v_config_file, vol->v_device_line,
+									"device", "device:%s:%s", h->name, vol->device);
 					}
 				}
 
@@ -111,6 +116,20 @@ struct d_host_info *find_host_info_by_name(struct d_resource* res, char *name)
 	return NULL;
 }
 
+static struct d_host_info *find_host_info_by_ha(struct d_resource* res, struct hname_address *ha)
+{
+	struct d_host_info *host;
+
+	if (!ha->faked_hostname)
+		return find_host_info_by_name(res, ha->name);
+
+	for_each_host(host, &res->all_hosts)
+		if (!strcmp(ha->name, names_to_str_c(&host->on_hosts, '_')))
+			return host;
+
+	return NULL;
+}
+
 static struct d_host_info *find_host_info_by_address(struct d_resource* res, struct d_address *address)
 {
 	struct d_host_info *host;
@@ -170,7 +189,7 @@ static void _set_host_info_in_host_address_pairs(struct d_resource *res,
 			/* The name will be used for nice comments only ... */
 			ha->name = strdup(names_to_str_c(&host_info->on_hosts, '_'));
 		} else {
-			host_info = find_host_info_by_name(res, ha->name);
+			host_info = find_host_info_by_ha(res, ha);
 		}
 		if (!host_info && !strcmp(ha->name, "_remote_host")) {
 			if (conn->peer)
@@ -406,7 +425,14 @@ static void set_peer_in_connection(struct d_resource* res, struct connection *co
 			if (conn->peer) {
 				host_info = conn->peer;
 			} else {
-				host_info = find_host_info_by_name(res, candidate->name);
+				host_info = find_host_info_by_ha(res, candidate);
+				if (!host_info) {
+					err("%s:%d: in connection in resource %s:\n"
+					    "\tCan not find host_info for %s\n",
+					    res->config_file, conn->config_line, res->name, candidate->name);
+					config_valid = 0;
+					return;
+				}
 				conn->peer = host_info;
 			}
 			path->peer_address = candidate->address.addr ? &candidate->address : &host_info->address;
@@ -444,7 +470,7 @@ bool peer_diskless(struct peer_device *peer_device)
 {
 	struct d_volume *vol;
 
-	vol = volume_by_vnr(&peer_device->connection->peer->volumes, peer_device->volume->vnr);
+	vol = volume_by_vnr(&peer_device->connection->peer->volumes, peer_device->vnr);
 	return vol->disk == NULL;
 }
 
@@ -491,9 +517,9 @@ void set_peer_in_resource(struct d_resource* res, int peer_required)
 		add_no_bitmap_opt(res);
 }
 
-void set_disk_in_res(struct d_resource *res)
+void set_stacked_disk_in_res(struct d_resource *res)
 {
-	struct d_host_info *host;
+	struct d_host_info *host, *lower_host;
 	struct d_volume *a, *b;
 
 	if (res->ignore)
@@ -503,41 +529,40 @@ void set_disk_in_res(struct d_resource *res)
 		if (!host->lower)
 			continue;
 
-		if (host->lower->ignore)
-			continue;
+		for_each_host(lower_host, &host->lower->all_hosts) {
+			check_volume_sets_equal(res, host, lower_host);
+			if (!config_valid)
+				/* don't even bother for broken config. */
+				continue;
 
-		check_volume_sets_equal(res, host, host->lower->me);
-		if (!config_valid)
-			/* don't even bother for broken config. */
-			continue;
-
-		/* volume lists are sorted on vnr */
-		a = STAILQ_FIRST(&host->volumes);
-		b = STAILQ_FIRST(&host->lower->me->volumes);
-		while (a) {
-			while (b && a->vnr > b->vnr) {
-				/* Lower resource has more volumes.
-				 * Probably unusual, but we decided
-				 * that it should be legal.
-				 * Skip those that do not match */
-				b = STAILQ_NEXT(b, link);
-			}
-			if (a && b && a->vnr == b->vnr) {
-				if (b->device)
-					m_asprintf(&a->disk, "%s", b->device);
-				else
-					m_asprintf(&a->disk, "/dev/drbd%u", b->device_minor);
-				/* stacked implicit volumes need internal meta data, too */
-				if (!a->meta_disk)
-					m_asprintf(&a->meta_disk, "internal");
-				if (!a->meta_index)
-					m_asprintf(&a->meta_index, "internal");
-				a = STAILQ_NEXT(a, link);
-				b = STAILQ_NEXT(b, link);
-			} else {
-				/* config_invalid should have been set
-				 * by check_volume_sets_equal */
-				assert(0);
+			/* volume lists are sorted on vnr */
+			a = STAILQ_FIRST(&host->volumes);
+			b = STAILQ_FIRST(&lower_host->volumes);
+			while (a) {
+				while (b && a->vnr > b->vnr) {
+					/* Lower resource has more volumes.
+					 * Probably unusual, but we decided
+					 * that it should be legal.
+					 * Skip those that do not match */
+					b = STAILQ_NEXT(b, link);
+				}
+				if (a && b && a->vnr == b->vnr) {
+					if (b->device)
+						m_asprintf(&a->disk, "%s", b->device);
+					else
+						m_asprintf(&a->disk, "/dev/drbd%u", b->device_minor);
+					/* stacked implicit volumes need internal meta data, too */
+					if (!a->meta_disk)
+						m_asprintf(&a->meta_disk, "internal");
+					if (!a->meta_index)
+						m_asprintf(&a->meta_index, "internal");
+					a = STAILQ_NEXT(a, link);
+					b = STAILQ_NEXT(b, link);
+				} else {
+					/* config_invalid should have been set
+					 * by check_volume_sets_equal */
+					assert(0);
+				}
 			}
 		}
 	}
@@ -573,21 +598,29 @@ static void inherit_volumes(struct volumes *from, struct d_host_info *host)
 			t = alloc_volume();
 			t->device_minor = -1;
 			t->vnr = s->vnr;
+			t->implicit = s->implicit;
+			t->v_config_file = s->v_config_file;
+			t->v_device_line = s->v_device_line;
+			t->v_disk_line = s->v_disk_line;
+			t->v_meta_disk_line = s->v_meta_disk_line;
 			insert_volume(&host->volumes, t);
 		}
 		if (!t->disk && s->disk) {
 			t->disk = strdup(s->disk);
 			STAILQ_FOREACH(h, &host->on_hosts, link)
-				check_uniq("disk", "disk:%s:%s", h->name, t->disk);
+				check_uniq_file_line(t->v_config_file, t->v_disk_line,
+					"disk", "disk:%s:%s", h->name, t->disk);
 		}
 		if (!t->device && s->device)
 			t->device = strdup(s->device);
 		if (t->device_minor == -1U && s->device_minor != -1U) {
 			t->device_minor = s->device_minor;
 			STAILQ_FOREACH(h, &host->on_hosts, link) {
-				check_uniq("device-minor", "device-minor:%s:%u", h->name, t->device_minor);
+				check_uniq_file_line(t->v_config_file, t->v_device_line,
+					"device-minor", "device-minor:%s:%u", h->name, t->device_minor);
 				if (t->device)
-					check_uniq("device", "device:%s:%s", h->name, t->device);
+					check_uniq_file_line(t->v_config_file, t->v_device_line,
+						"device", "device:%s:%s", h->name, t->device);
 			}
 		}
 		if (!t->meta_disk && s->meta_disk) {
@@ -640,7 +673,8 @@ static void check_meta_disk(struct d_volume *vol, struct d_host_info *host)
 	if (strcmp(vol->meta_disk, "internal") != 0) {
 		/* index either some number, or "flexible" */
 		STAILQ_FOREACH(h, &host->on_hosts, link)
-			check_uniq("meta-disk", "%s:%s[%s]", h->name, vol->meta_disk, vol->meta_index);
+			check_uniq_file_line(vol->v_config_file, vol->v_meta_disk_line,
+				"meta-disk", "%s:%s[%s]", h->name, vol->meta_disk, vol->meta_index);
 	}
 }
 
@@ -684,6 +718,15 @@ static void check_volume_sets_equal(struct d_resource *res, struct d_host_info *
 			b = STAILQ_NEXT(b, link);
 		}
 		if (a && b && a->vnr == b->vnr) {
+			if (a->implicit != b->implicit) {
+				err("%s:%d: in resource %s, on %s resp. %s: volume %d must not be implicit on one but not the other\n",
+				    config_file, line, res->name,
+				    names_to_str(&host1->on_hosts),
+				    compare_stacked ? host1->lower->name : names_to_str(&host2->on_hosts),
+				    a->vnr);
+				config_valid = 0;
+			}
+
 			a = STAILQ_NEXT(a, link);
 			b = STAILQ_NEXT(b, link);
 		}
@@ -771,6 +814,14 @@ static struct d_host_info *find_host_info_or_invalid(struct d_resource *res, cha
 		    "There is no 'on' section for hostname '%s' named in the connection-mesh\n",
 		    res->config_file, res->start_line, res->name, name);
 		config_valid = 0;
+	} else if (host_info->proxy_compat_only && config_valid) {
+		err("%s:%d: in resource %s:\n\t"
+		    "There is a drbd-8.x proxy config syntax within an 'on' section and you are using\n\t"
+		    "the 'mesh' keyword from drbd-9.x syntax. That does not work together.\n\n\t"
+		    "Use an explicit connection to configure a proxy and omit that pair of hosts\n\t"
+		    "from the mesh.\n",
+		    res->config_file, host_info->config_line, res->name);
+		config_valid = 0;
 	}
 
 	return host_info;
@@ -804,7 +855,7 @@ static void create_connections_from_mesh(struct d_resource *res, struct mesh *me
 			path->implicit = 1;
 			insert_tail(&conn->paths, path);
 
-			expand_opts(&mesh->net_options, &conn->net_options);
+			expand_opts(res, &show_net_options_ctx, &mesh->net_options, &conn->net_options);
 
 			ha = alloc_hname_address();
 			ha->host_info = hi1;
@@ -823,22 +874,40 @@ static void create_connections_from_mesh(struct d_resource *res, struct mesh *me
 	}
 }
 
+// Returns 0 if the addresses, ports, address families match.
+// Otherwise, returns the result of the comparison of those
+// strings that mismatch.
 int addresses_cmp(struct d_address *addr1, struct d_address *addr2)
 {
-	int ret;
+	bool ips_match = false;
+	if (strcasecmp(addr1->af, IPV4_STR) == 0 &&
+	    strcasecmp(addr2->af, IPV4_STR) == 0) {
+		ips_match = ipv4_addresses_match(addr1->addr, addr2->addr);
+	} else if (strcasecmp(addr1->af, IPV6_STR) == 0 &&
+	           strcasecmp(addr2->af, IPV6_STR) == 0) {
+		ips_match = ipv6_addresses_match(addr1->addr, addr2->addr);
+	}
 
-	if ((ret = strcmp(addr1->addr, addr2->addr)))
-		return ret;
+	int result = ips_match ? 0 : strcmp(addr1->addr, addr2->addr);
 
-	if ((ret = strcmp(addr1->port, addr2->port)))
-		return ret;
+	// If the addresses did not match, emulate the old behavior,
+	// because this function is used for sorting tree entries elsewhere
+	if (result == 0) {
+		// Peer addresses match, order lexically by port
+		result = strcmp(addr1->port, addr2->port);
+		if (result == 0) {
+			// Peer addresses and ports match,
+			// order lexically by address family
+			result = strcmp(addr1->af, addr2->af);
+		}
+	}
 
-	return strcmp(addr1->af, addr2->af);
+	return result;
 }
 
 bool addresses_equal(struct d_address *addr1, struct d_address *addr2)
 {
-	return !addresses_cmp(addr1, addr2);
+	return addresses_cmp(addr1, addr2) == 0;
 }
 
 struct addrtree_entry {
@@ -974,18 +1043,13 @@ static void must_have_two_hosts(struct d_resource *res, struct connection *conn)
 		_must_have_two_hosts(res, path);
 }
 
-struct peer_device *find_peer_device(struct d_host_info *host, struct connection *conn, int vnr)
+struct peer_device *find_peer_device(struct connection *conn, int vnr)
 {
 	struct peer_device *peer_device;
-	struct d_volume *vol;
 
 	STAILQ_FOREACH(peer_device, &conn->peer_devices, connection_link) {
-		if (peer_device->vnr == vnr) {
-			for_each_volume(vol, &host->volumes) {
-				if (vol == peer_device->volume)
-					return peer_device;
-			}
-		}
+		if (peer_device->vnr == vnr)
+			return peer_device;
 	}
 
 	return NULL;
@@ -1006,39 +1070,32 @@ static void fixup_peer_devices(struct d_resource *res)
 		if (!some_path)
 			continue;
 
-		some_host = STAILQ_FIRST(&some_path->hname_address_pairs)->host_info;
-
 		STAILQ_FOREACH(peer_device, &conn->peer_devices, connection_link) {
-
-			vol = volume_by_vnr(&some_host->volumes, peer_device->vnr);
-			if (!vol) {
-				err("%s:%d: Resource %s: There is a reference to a volume %d that"
-				    "is not known in this resource\n",
-				    res->config_file, peer_device->config_line, res->name,
-				    peer_device->vnr);
-				config_valid = 0;
+			STAILQ_FOREACH(ha, &some_path->hname_address_pairs, link) {
+				struct d_host_info *host = ha->host_info;
+				if (!strcmp("_remote_host", ha->name)) /* && PARSE_FOR_ADJUST */
+					continue; /* no on section for _remote_host in show output! */
+				vol = volume_by_vnr(&host->volumes, peer_device->vnr);
+				if (!vol) {
+					err("%s:%d: Resource %s: There is a reference to a volume %d that "
+					    "is not known in this resource on host %s\n",
+					    res->config_file, peer_device->config_line, res->name,
+					    peer_device->vnr, ha->name);
+					config_valid = 0;
+				}
 			}
-			peer_device->volume = vol;
-			STAILQ_INSERT_TAIL(&vol->peer_devices, peer_device, volume_link);
 		}
 
-		/* Take the first path, iterate over hname_address_pairs, take the host_info.
-		   this is the way to get hold of the two hosts of a connection. */
-		STAILQ_FOREACH(ha, &some_path->hname_address_pairs, link) {
-			struct d_host_info *host = ha->host_info;
-
-			for_each_volume(vol, &host->volumes) {
-				peer_device = find_peer_device(host, conn, vol->vnr);
-				if (peer_device)
-					continue;
-				peer_device = alloc_peer_device();
-				peer_device->vnr = vol->vnr;
-				peer_device->implicit = 1;
-				peer_device->connection = conn;
-				peer_device->volume = vol;
-				STAILQ_INSERT_TAIL(&conn->peer_devices, peer_device, connection_link);
-				STAILQ_INSERT_TAIL(&vol->peer_devices, peer_device, volume_link);
-			}
+		some_host = STAILQ_FIRST(&some_path->hname_address_pairs)->host_info;
+		for_each_volume(vol, &some_host->volumes) {
+			peer_device = find_peer_device(conn, vol->vnr);
+			if (peer_device)
+				continue;
+			peer_device = alloc_peer_device();
+			peer_device->vnr = vol->vnr;
+			peer_device->implicit = 1;
+			peer_device->connection = conn;
+			STAILQ_INSERT_TAIL(&conn->peer_devices, peer_device, connection_link);
 		}
 	}
 }
@@ -1051,18 +1108,31 @@ void post_parse(struct resources *resources, enum pp_flags flags)
 	/* inherit volumes from resource level into the d_host_info objects */
 	for_each_resource(res, resources) {
 		struct d_host_info *host;
+		bool any_implicit = false;
+		bool any_non_zero_vnr = false;
 		for_each_host(host, &res->all_hosts) {
 			struct d_volume *vol;
+
 			inherit_volumes(&res->volumes, host);
 
-			for_each_volume(vol, &host->volumes)
+			for_each_volume(vol, &host->volumes) {
+				any_implicit |= vol->implicit;
+				any_non_zero_vnr |= vol->vnr != 0;
+
 				check_meta_disk(vol, host);
+			}
 
 			if (host->require_minor)
 				check_volumes_complete(res, host);
 		}
 
 		check_volumes_hosts(res);
+
+		if (any_implicit && any_non_zero_vnr) {
+			err("%s:%d: in resource %s: you must not mix implicit any explicit volumes\n",
+			    config_file, line, res->name);
+			config_valid = 0;
+		}
 	}
 
 	for_each_resource(res, resources)
@@ -1110,21 +1180,40 @@ void post_parse(struct resources *resources, enum pp_flags flags)
 	// Needs "me" set already
 	for_each_resource(res, resources)
 		if (res->stacked_on_one)
-			set_disk_in_res(res);
+			set_stacked_disk_in_res(res);
 
 	for_each_resource(res, resources)
 		fixup_peer_devices(res);
 }
 
-static void expand_opts(struct options *common, struct options *options)
+static void expand_opts(struct d_resource *res, struct context_def *oc, struct options *common, struct options *options)
 {
-	struct d_option *option, *new_option;
+	struct d_option *option, *new_option, *existing_option;
 
 	STAILQ_FOREACH(option, common, link) {
-		if (!find_opt(options, option->name)) {
+		existing_option = find_opt(options, option->name);
+		if (!existing_option) {
 			new_option = new_opt(strdup(option->name),
 					     option->value ? strdup(option->value) : NULL);
+			new_option->inherited = true;
 			insert_head(options, new_option);
+		} else if (existing_option->inherited && oc != &wildcard_ctx) {
+			if (!is_equal(oc, existing_option, option)) {
+				err("%s:%d: in resource %s, "
+				    "ambiguous inheritance for option \"%s\".\n"
+				    "should be \"%s\" and \"%s\" at the same time\n.",
+				    res->config_file, res->start_line, res->name,
+				    option->name, existing_option->value, option->value);
+				config_valid = 0;
+			}
+			/* else {
+				err("%s:%d: WARNING: in resource %s, "
+				    "multiple inheritance for option \"%s\".\n"
+				    "with same value\n.",
+				    res->config_file, res->start_line, res->name,
+				    option->name, existing_option->value, option->value);
+                        }
+			*/
 		}
 	}
 }
@@ -1162,26 +1251,26 @@ void expand_common(void)
 			template = common;
 
 		if (template) {
-			expand_opts(&template->net_options, &res->net_options);
-			expand_opts(&template->disk_options, &res->disk_options);
-			expand_opts(&template->pd_options, &res->pd_options);
-			expand_opts(&template->startup_options, &res->startup_options);
-			expand_opts(&template->proxy_options, &res->proxy_options);
-			expand_opts(&template->handlers, &res->handlers);
-			expand_opts(&template->res_options, &res->res_options);
+			expand_opts(res, &show_net_options_ctx, &template->net_options, &res->net_options);
+			expand_opts(res, &disk_options_ctx, &template->disk_options, &res->disk_options);
+			expand_opts(res, &device_options_ctx, &template->pd_options, &res->pd_options);
+			expand_opts(res, &startup_options_ctx, &template->startup_options, &res->startup_options);
+			expand_opts(res, &proxy_options_ctx, &template->proxy_options, &res->proxy_options);
+			expand_opts(res, &handlers_ctx, &template->handlers, &res->handlers);
+			expand_opts(res, &resource_options_ctx, &template->res_options, &res->res_options);
 
 			if (template->stacked_timeouts)
 				res->stacked_timeouts = 1;
 
-			expand_opts(&template->proxy_plugins, &res->proxy_plugins);
+			expand_opts(res, &wildcard_ctx, &template->proxy_plugins, &res->proxy_plugins);
 		}
 
 		/* now that common disk options (if any) have been propagated to the
 		 * resource level, further propagate them to the volume level. */
 		for_each_host(h, &res->all_hosts) {
 			for_each_volume(vol, &h->volumes) {
-				expand_opts(&res->disk_options, &vol->disk_options);
-				expand_opts(&res->pd_options, &vol->pd_options);
+				expand_opts(res, &disk_options_ctx, &res->disk_options, &vol->disk_options);
+				expand_opts(res, &peer_device_options_ctx, &res->pd_options, &vol->pd_options);
 			}
 		}
 
@@ -1189,14 +1278,14 @@ void expand_common(void)
 		for_each_volume(vol, &res->volumes) {
 			for_each_host(h, &res->all_hosts) {
 				host_vol = volume_by_vnr(&h->volumes, vol->vnr);
-				expand_opts(&vol->disk_options, &host_vol->disk_options);
-				expand_opts(&vol->pd_options, &host_vol->pd_options);
+				expand_opts(res, &disk_options_ctx, &vol->disk_options, &host_vol->disk_options);
+				expand_opts(res, &peer_device_options_ctx, &vol->pd_options, &host_vol->pd_options);
 			}
 		}
 
 		/* inherit network options from resource objects into connection objects */
 		for_each_connection(conn, &res->connections)
-			expand_opts(&res->net_options, &conn->net_options);
+			expand_opts(res, &show_net_options_ctx, &res->net_options, &conn->net_options);
 
 		/* inherit proxy options from resource to the proxies in the connections */
 		for_each_connection(conn, &res->connections) {
@@ -1207,8 +1296,8 @@ void expand_common(void)
 					if (!ha->proxy)
 						continue;
 
-					expand_opts(&res->proxy_options, &ha->proxy->options);
-					expand_opts(&res->proxy_plugins, &ha->proxy->plugins);
+					expand_opts(res, &proxy_options_ctx, &res->proxy_options, &ha->proxy->options);
+					expand_opts(res, &wildcard_ctx, &res->proxy_plugins, &ha->proxy->plugins);
 				}
 			}
 		}
@@ -1217,9 +1306,23 @@ void expand_common(void)
 		   tie the peer_device options from the volume to peer_devices */
 		for_each_connection(conn, &res->connections) {
 			struct peer_device *peer_device;
-			STAILQ_FOREACH(peer_device, &conn->peer_devices, connection_link) {
-				expand_opts(&conn->pd_options, &peer_device->pd_options);
-				expand_opts(&peer_device->volume->pd_options, &peer_device->pd_options);
+			struct hname_address *ha;
+			struct path *some_path;
+
+			STAILQ_FOREACH(peer_device, &conn->peer_devices, connection_link)
+				expand_opts(res, &peer_device_options_ctx, &conn->pd_options, &peer_device->pd_options);
+
+			some_path = STAILQ_FIRST(&conn->paths);
+			if (!some_path)
+				continue;
+
+			STAILQ_FOREACH(ha, &some_path->hname_address_pairs, link) {
+				h = ha->host_info;
+				for_each_volume(vol, &h->volumes) {
+					peer_device = find_peer_device(conn, vol->vnr);
+
+					expand_opts(res, &peer_device_options_ctx, &vol->pd_options, &peer_device->pd_options);
+				}
 			}
 		}
 	}
